/**
 * Abstraction for threading to allow porting.
 * Currently using either pthread or C++ 11 threads.
 */

#ifndef PLS_THREAD_H
#define PLS_THREAD_H

// platform specific includes
#ifdef PLS_USING_PTHREADS
    #include <pthread.h>
#elif PLS_USING_CPP_THREADS
    #include <thread>
#else
    #error "Please configure exactly one threading library!"
#endif

namespace pls {
    namespace internal {
        namespace base {
            using thread_entrypoint = void();

            // Thread local storage support
#ifdef PLS_USING_PTHREADS
            pthread_key_t thr_id_key;
            bool thr_id_key_created = false;
            // forward declaration
            template<typename T>
            void* start_pthread_internal(void*);
#endif
#ifdef PLS_USING_CPP_THREADS
            thread_local void* local_thread;
#endif


            template<typename T>
            class thread {
            private:
                // Handle to the native thread used
#ifdef PLS_USING_PTHREADS
                friend void* start_pthread_internal<T>(void*);

                pthread_t pthread_thread_;
                thread_entrypoint* entry_function_;

                thread(thread_entrypoint entry_function, T local_object):
                pthread_thread_(),
                entry_function_(entry_function),
                local_object_(local_object) {
                    if (!thr_id_key_created) {
                        pthread_key_create(&thr_id_key, nullptr);
                        thr_id_key_created = true;
                    }

                    pthread_create(&pthread_thread_, nullptr, start_pthread_internal<T>, (void *)(this));
                }
#endif
#ifdef PLS_USING_CPP_THREADS
                std::thread std_thread_;

                thread(thread_entrypoint entry_function, T local_object):
                local_object_(local_object),
                std_thread_([=](){
                    local_thread = this;
                    entry_function();
                }) {};
#endif
            public:
                T local_object_;

                /**
                 * Creates and starts a thread.
                 * NOT thread safe, best only use from one main thread managing the runtime!
                 *
                 * @param entry_function The entry function to run on the thread
                 * @param T local_object
                 *
                 * @return a handle to the newly created thread
                 */
                static thread start(thread_entrypoint entry_function, T local_object) {
                    return thread(entry_function, local_object);
                }

                void join() {
#ifdef PLS_USING_PTHREADS
                    pthread_join(pthread_thread_, nullptr);
#endif
#ifdef PLS_USING_CPP_THREADS
                    std_thread_.join();
#endif
                }

                static void yield() {
#ifdef PLS_USING_PTHREADS
                    pthread_yield();
#endif
#ifdef PLS_USING_CPP_THREADS
                    std::this_thread::yield();
#endif
                }

                static thread* get_current() {
#ifdef PLS_USING_PTHREADS
                    return (thread*) pthread_getspecific(thr_id_key);
#endif
#ifdef PLS_USING_CPP_THREADS
                    return (thread*) local_thread;
#endif
                }

                // make object move only
                thread(thread&&) noexcept = default;
                thread& operator=(thread&&) noexcept = default;

                thread(const thread&) = delete;
                thread& operator=(const thread&) = delete;
            };

#ifdef PLS_USING_PTHREADS
            template<typename T>
            void* start_pthread_internal(void* thread_pointer) {
                auto* my_thread = (thread<T>*)thread_pointer;
                pthread_setspecific(thr_id_key, thread_pointer);
                my_thread->entry_function_();
                pthread_exit(nullptr);
            }
#endif
        }
    }

}

#endif //PLS_THREAD_H
